@@ -3,7 +3,7 @@
 # Programmer friendly subprocess wrapper.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: May 23, 2015
+# Last Change: May 25, 2015
 # URL: https://executor.readthedocs.org
 
 """
@@ -47,28 +47,50 @@ Classes and functions
 import logging
 import os
 import pipes
+import signal
 import subprocess
 import tempfile
 
 # Modules included in our package.
-from executor.writable_property import override_properties, writable_property
-
-# We need to know whether standard input should be encoded or not. The way to
-# check depends on whether we are running on Python 2.x or Python 3.x (both of
-# which we want to support).
-try:
-    # On Python 2.x this does nothing, on Python 3.x it will raise NameError.
-    unicode = unicode
-except NameError:
-    # Define an unambiguous alias for our intention on Python 3.x.
-    unicode = str
+from executor.compat import str
+from executor.property_manager import (
+    mutable_property,
+    PropertyManager,
+    required_property,
+)
 
 # Semi-standard module versioning.
-__version__ = '2.1'
+__version__ = '3.0'
 
 # Initialize a logger.
 logger = logging.getLogger(__name__)
 
+DEFAULT_SHELL = 'bash'
+"""
+The default shell used to evaluate shell expressions (a string).
+
+This variable isn't based on the ``$SHELL`` environment variable because:
+
+1. Shells like ``sh``, ``dash``, ``bash`` and ``zsh`` all have their own
+   subtly incompatible semantics.
+2. People regularly use shells like ``fish`` as their default login shell :-).
+
+At an interactive prompt this is no problem (advanced users have obviously
+learned to context switch) but when you're writing source code the last thing
+you want to worry about is which shell is going to evaluate your commands! The
+:mod:`executor` package expects this shell to support the following features:
+
+- The ``-c`` option to evaluate a shell command provided as a command line
+  argument.
+
+- The ``-`` argument to instruct the shell to read shell commands from its
+  standard input stream and evaluate those.
+
+Apart from these two things nothing else is expected from the default shell so
+you're free to customize it if you really want to write your shell commands in
+``fish`` or ``zsh`` syntax :-).
+"""
+
 
 def execute(*command, **options):
     """
@@ -124,7 +146,7 @@ def execute(*command, **options):
         self.check_errors()
       File "executor/__init__.py", line 568, in check_errors
         raise ExternalCommandFailed(self)
-    executor.ExternalCommandFailed: External command failed with exit code 1! (command: bash -c false)
+    executor.ExternalCommandFailed: External command failed with exit code 1! (command: false)
 
     The exceptions raised by :func:`execute()` expose
     :attr:`~ExternalCommandFailed.command` and
@@ -148,7 +170,7 @@ def execute(*command, **options):
             return cmd.returncode == 0
 
 
-class ExternalCommand(object):
+class ExternalCommand(PropertyManager):
 
     """
     Programmer friendly :class:`subprocess.Popen` wrapper.
@@ -162,25 +184,44 @@ class ExternalCommand(object):
     Because the :class:`ExternalCommand` class has a lot of properties and
     methods here is a summary:
 
-    - The writable properties :attr:`async`, :attr:`capture`, :attr:`check`,
-      :attr:`command`, :attr:`directory`, :attr:`encoding`,
-      :attr:`environment`, :attr:`fakeroot`, :attr:`input`, :attr:`logger`,
-      :attr:`silent` and :attr:`sudo` allow you to configure how the external
-      command will be run (before it is started).
+    **Writable properties**
+     The :attr:`async`, :attr:`capture`, :attr:`check`, :attr:`command`,
+     :attr:`directory`, :attr:`encoding`, :attr:`environment`,
+     :attr:`fakeroot`, :attr:`input`, :attr:`logger`, :attr:`silent` and
+     :attr:`sudo` properties allow you to configure how the external command
+     will be run (before it is started).
 
-    - The read only properties :attr:`command_line`, :attr:`encoded_input`,
-      :attr:`is_finished`, :attr:`is_running`, :attr:`output`,
-      :attr:`returncode`, :attr:`stdout` and :attr:`was_started` allow you to
-      inspect if and how the external command was started, what its current
-      status is and what its output is.
+    **Computed properties**
+     The :attr:`command_line`, :attr:`encoded_input`, :attr:`is_finished`,
+     :attr:`is_running`, :attr:`is_terminated`, :attr:`output`,
+     :attr:`returncode`, :attr:`stdout` and :attr:`was_started` properties
+     allow you to inspect if and how the external command was started, what its
+     current status is and what its output is.
 
-    - The public methods :func:`start()` and :func:`wait()` enable you to start
-      external commands and wait for them to finish.
+    **Public methods**
+     The public methods :func:`start()`, :func:`wait()` and :func:`terminate()`
+     enable you to start external commands, wait for them to finish and
+     terminate them if they take too long.
 
-    - The internal methods :func:`check_errors()`, :func:`load_output()` and
-      :func:`cleanup()` are used by :func:`start()` and :func:`wait()` so
-      unless you're reimplementing one of those two methods you probably don't
-      need these internal methods.
+    **Internal methods**
+     The internal methods :func:`check_errors()`, :func:`load_output()` and
+     :func:`cleanup()` are used by :func:`start()`, :func:`wait()` and
+     :func:`terminate()` so unless you're reimplementing one of those methods
+     you probably don't need these internal methods.
+
+    **Context manager**
+      :class:`ExternalCommand` objects can be used as context managers by using
+      the :keyword:`with` statement:
+
+      - When the scope of the :keyword:`with` statement starts the
+        :func:`start()` method is called (if the external command
+        isn't already running).
+      - When the scope of the :keyword:`with` statement ends
+        :func:`terminate()` is called if the command is still running. The
+        :func:`load_output()` and :func:`cleanup()` functions are used to
+        cleanup after the external command. If an exception isn't already being
+        raised :func:`check_errors()` is called to make sure the external
+        command succeeded.
     """
 
     def __init__(self, *command, **options):
@@ -201,9 +242,12 @@ class ExternalCommand(object):
         :func:`wait()`.
         """
         # Store the command and its arguments.
-        self.command = list(command)
-        # Use any keyword arguments to override defaults.
-        override_properties(self, **options)
+        command = list(command)
+        if not command:
+            raise TypeError("Please provide a command to execute!")
+        self.command = command
+        # Set properties based on keyword arguments.
+        super(ExternalCommand, self).__init__(**options)
         # Initialize instance variables.
         self.cached_stdout = None
         self.input_file = None
@@ -211,7 +255,7 @@ class ExternalCommand(object):
         self.output_file = None
         self.subprocess = None
 
-    @writable_property
+    @mutable_property
     def async(self):
         """
         If this option is :data:`True` (not the default) preparations are made
@@ -241,7 +285,7 @@ class ExternalCommand(object):
         """
         return False
 
-    @writable_property
+    @mutable_property
     def capture(self):
         """
         If this option is :data:`True` (not the default) the standard output of
@@ -253,7 +297,7 @@ class ExternalCommand(object):
         """
         return False
 
-    @writable_property
+    @mutable_property
     def check(self):
         """
         If this option is :data:`True` (the default) and the external command
@@ -263,12 +307,9 @@ class ExternalCommand(object):
         """
         return True
 
-    @writable_property
+    @required_property
     def command(self):
-        """
-        A list of strings with the command to execute.
-        """
-        return []
+        """A list of strings with the command to execute."""
 
     @property
     def command_line(self):
@@ -278,8 +319,9 @@ class ExternalCommand(object):
         :attr:`command` according to the following rules:
 
         - If :attr:`command` contains a single string it is assumed to be a
-          shell command and run using ``bash -c '...'`` which means constructs
-          like semicolons, ampersands and pipes can be used (and all the usual
+          shell command and run using ``bash -c '...'`` (assuming you haven't
+          changed :data:`DEFAULT_SHELL`) which means constructs like
+          semicolons, ampersands and pipes can be used (and all the usual
           caveats apply :-).
 
         - If :attr:`fakeroot` or :attr:`sudo` is set the respective command
@@ -287,8 +329,8 @@ class ExternalCommand(object):
         """
         command_line = list(self.command)
         if len(command_line) == 1:
-            command_line = ['bash', '-c'] + command_line
-        if (self.fakeroot or self.sudo) and os.getuid() != 0:
+            command_line = [DEFAULT_SHELL, '-c'] + command_line
+        if (self.fakeroot or self.sudo) and not self.have_superuser_privileges:
             if self.sudo:
                 # Superuser privileges requested by caller.
                 command_line = ['sudo'] + command_line
@@ -300,7 +342,7 @@ class ExternalCommand(object):
                 command_line = ['sudo'] + command_line
         return command_line
 
-    @writable_property
+    @mutable_property
     def directory(self):
         """
         The working directory for the external command (a string). Defaults to
@@ -312,14 +354,14 @@ class ExternalCommand(object):
     def encoded_input(self):
         """
         The value of :attr:`input` encoded using :attr:`encoding`. This is a
-        :class:`python2:str` object (in Python 2.x) or a :class:`python3:bytes`
-        object (in Python 3.x).
+        :class:`python2:str` object (in Python 2) or a :class:`python3:bytes`
+        object (in Python 3).
         """
         return (self.input.encode(self.encoding)
-                if isinstance(self.input, unicode)
+                if isinstance(self.input, str)
                 else self.input)
 
-    @writable_property
+    @mutable_property
     def encoding(self):
         """
         The character encoding of standard input and standard output (a string,
@@ -328,7 +370,7 @@ class ExternalCommand(object):
         """
         return 'UTF-8'
 
-    @writable_property
+    @mutable_property
     def environment(self):
         """
         A dictionary of environment variables for the external command.
@@ -339,40 +381,63 @@ class ExternalCommand(object):
         """
         return {}
 
-    @writable_property
+    @property
+    def error_message(self):
+        """
+        A string describing how the external command failed or :data:`None`.
+        """
+        if self.error_type is ExternalCommandFailed:
+            text = "External command failed with exit code %s! (command: %s)"
+            return text % (self.returncode, quote(self.command_line))
+
+    @property
+    def error_type(self):
+        """
+        :exc:`ExternalCommandFailed` when :attr:`returncode` is set and not
+        zero, :data:`None` otherwise.
+        """
+        if self.returncode is not None:
+            if self.returncode != 0:
+                return ExternalCommandFailed
+
+    @mutable_property
     def fakeroot(self):
         """
         If this option is :data:`True` (not the default) and the current
         process doesn't have `superuser privileges`_ the external command is
         run with ``fakeroot``. If the ``fakeroot`` program is not installed a
         fall back to ``sudo`` is performed.
+
+        .. _superuser privileges: http://en.wikipedia.org/wiki/Superuser#Unix_and_Unix-like
         """
         return False
 
-    @writable_property
+    @property
+    def have_superuser_privileges(self):
+        """
+        :data:`True` if running with `superuser privileges`_, :data:`False`
+        otherwise. Used by :attr:`command_line` to decide whether
+        :attr:`fakeroot` or :attr:`sudo` needs to be used.
+        """
+        return os.getuid() == 0
+
+    @mutable_property
     def input(self):
         """
-        The input to feed to the external command on the standard input stream.
+        The input to feed to the external command on the standard input stream
+        (defaults to :data:`None`).
 
-        When you provide a :func:`python2:unicode` object (in Python 2.x) or a
-        :class:`python3:str` object (in Python 3.x) as input it will be encoded
+        When you provide a :func:`python2:unicode` object (in Python 2) or a
+        :class:`python3:str` object (in Python 3) as input it will be encoded
         using :attr:`encoding`. To avoid the automatic conversion you can
-        simply pass a :class:`python2:str` object (in Python 2.x) or a
-        :class:`python3:bytes` object (in Python 3.x).
+        simply pass a :class:`python2:str` object (in Python 2) or a
+        :class:`python3:bytes` object (in Python 3).
 
         The conversion logic is implemented in the :attr:`encoded_input`
         attribute.
         """
         return None
 
-    @property
-    def is_running(self):
-        """
-        :data:`True` while the external command is running, :data:`False` when
-        the external command hasn't been started yet or has already finished.
-        """
-        return self.subprocess.poll() is None if self.subprocess else False
-
     @property
     def is_finished(self):
         """
@@ -382,7 +447,24 @@ class ExternalCommand(object):
         """
         return self.subprocess.poll() is not None if self.subprocess else False
 
-    @writable_property
+    @property
+    def is_running(self):
+        """
+        :data:`True` while the external command is running, :data:`False` when
+        the external command hasn't been started yet or has already finished.
+        """
+        return self.subprocess.poll() is None if self.subprocess else False
+
+    @property
+    def is_terminated(self):
+        """
+        :data:`True` if the external command was terminated using
+        :data:`signal.SIGTERM` (e.g. by :func:`terminate()`),
+        :data:`False` otherwise.
+        """
+        return abs(self.returncode) == signal.SIGTERM if self.returncode and self.returncode < 0 else False
+
+    @mutable_property
     def logger(self):
         """
         The :class:`logging.Logger` object to use.
@@ -399,8 +481,8 @@ class ExternalCommand(object):
     def output(self):
         r"""
         The value of :attr:`stdout` decoded using :attr:`encoding`. This is a
-        :func:`python2:unicode` object (in Python 2.x) or a
-        :class:`python3:str` object (in Python 3.x).
+        :func:`python2:unicode` object (in Python 2) or a :class:`python3:str`
+        object (in Python 3).
 
         This is only available when :attr:`capture` is :data:`True`. If
         :attr:`capture` is not :data:`True` then :attr:`output` will be
@@ -438,7 +520,7 @@ class ExternalCommand(object):
         """
         return self.subprocess.poll() if self.subprocess else None
 
-    @writable_property
+    @mutable_property
     def silent(self):
         """
         If this is :data:`True` (not the default) any output of the external
@@ -454,8 +536,8 @@ class ExternalCommand(object):
     def stdout(self):
         """
         The output of the external command on its standard output stream, a
-        :class:`python2:str` object (in Python 2.x) or a :class:`python3:bytes`
-        object (in Python 3.x).
+        :class:`python2:str` object (in Python 2) or a :class:`python3:bytes`
+        object (in Python 3).
 
         This is only available when :attr:`capture` is :data:`True`. If
         :attr:`capture` is not :data:`True` then :attr:`stdout` will be
@@ -468,15 +550,13 @@ class ExternalCommand(object):
             self.load_output()
         return self.cached_stdout
 
-    @writable_property
+    @mutable_property
     def sudo(self):
         """
         If this option is :data:`True` (not the default) and the current
         process doesn't have `superuser privileges`_ the external command is
         run with ``sudo`` to ensure that the external command runs with
         superuser privileges.
-
-        .. _superuser privileges: http://en.wikipedia.org/wiki/Superuser#Unix_and_Unix-like
         """
         return False
 
@@ -532,7 +612,7 @@ class ExternalCommand(object):
         if self.capture:
             if self.async:
                 fd, self.output_file = tempfile.mkstemp(prefix='executor-', suffix='-output.txt')
-                logger.debug("Capturing external command output in temporary file %s ..", self.output_file)
+                self.logger.debug("Capturing external command output in temporary file %s ..", self.output_file)
                 kw['stdout'] = fd
             else:
                 kw['stdout'] = subprocess.PIPE
@@ -546,10 +626,13 @@ class ExternalCommand(object):
             self.cached_stdout, stderr = self.subprocess.communicate(input=self.encoded_input)
             self.wait()
 
-    def wait(self):
+    def wait(self, check=None):
         """
         Wait for the external command to finish.
 
+        :param check: Override the value of :attr:`check` for the duration of
+                      this call to :func:`wait()`. Defaults to :data:`None`
+                      which means :attr:`check` is not overridden.
         :raises: :exc:`ExternalCommandFailed` when :attr:`check` is
                  :data:`True`, :attr:`async` is :data:`True` and the external
                  command exits with a nonzero status code.
@@ -573,7 +656,23 @@ class ExternalCommand(object):
             self.subprocess.wait()
         self.load_output()
         self.cleanup()
-        self.check_errors()
+        self.check_errors(check=check)
+
+    def terminate(self):
+        """
+        Terminate a running process using :func:`subprocess.Popen.terminate()`.
+
+        Calls :func:`wait()` after terminating the process so that the external
+        command's output is loaded and temporary resources are cleaned up. The
+        value of :attr:`check` is overridden to :data:`False` during the call
+        to :func:`terminate()` (if you're terminating an external command you
+        know the return code isn't going to be zero so there's no point in
+        raising an exception about it).
+        """
+        if self.is_running:
+            self.logger.debug("Terminating external command: %s", quote(self.command_line))
+            self.subprocess.terminate()
+            self.wait(check=False)
 
     def load_output(self):
         """
@@ -603,51 +702,43 @@ class ExternalCommand(object):
             self.null_device.close()
             self.null_device = None
 
-    def check_errors(self):
+    def check_errors(self, check=None):
         """
-        Raise :exc:`ExternalCommandFailed` when :attr:`check` is set and the
-        external command ended with a nonzero exit code.
+        Raise :attr:`error_type` when :attr:`check` is set and the external
+        command failed.
+
+        :param check: Override the value of :attr:`check` for the duration of
+                      this call. Defaults to :data:`None` which means
+                      :attr:`check` is not overridden.
+        :raises: :attr:`error_type` when :attr:`check` is set and
+                 :attr:`error_type` is not :data:`None`.
 
         This internal method is used by :func:`start()` and :func:`wait()` to
         make sure that failing external commands don't go unnoticed.
         """
-        if self.check and self.returncode != 0:
-            raise ExternalCommandFailed(self)
+        if (check if check is not None else self.check) and self.error_type is not None:
+            raise self.error_type(self)
 
-    def __repr__(self):
+    def __enter__(self):
+        """Start the external command if it hasn't already been started."""
+        if not self.was_started:
+            self.start()
+        return self
+
+    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
         """
-        Report a user friendly representation of a :class:`ExternalCommand` object.
+        Terminate the external command if it is still running, cleanup after
+        the external command and check for errors.
         """
-        fields = [
-            "async=%r" % self.async,
-            "capture=%r" % self.capture,
-            "check=%r" % self.check,
-            "command=%r" % self.command,
-            "encoding=%r" % self.encoding,
-            "is_finished=%r" % self.is_finished,
-            "is_running=%r" % self.is_running,
-            "was_started=%r" % self.was_started,
-        ]
-        # The following fields are only included when they're useful.
-        if self.command_line != self.command:
-            fields.append("command_line=%r" % self.command_line)
-        if self.directory != os.curdir:
-            fields.append("directory=%r" % self.directory)
-        if self.environment:
-            fields.append("environment=%r" % self.environment)
-        if self.fakeroot:
-            fields.append("fakeroot=%r" % self.fakeroot)
-        if self.input:
-            fields.append("input=%r" % self.input)
-        if self.output:
-            fields.append("output=%r" % self.output)
-        if self.returncode is not None:
-            fields.append("returncode=%r" % self.returncode)
-        if self.silent:
-            fields.append("silent=%r" % self.silent)
-        if self.sudo:
-            fields.append("sudo=%r" % self.sudo)
-        return "%s(%s)" % (self.__class__.__name__, ", ".join(sorted(fields)))
+        if self.was_started:
+            if self.is_running:
+                self.terminate()
+            self.load_output()
+            self.cleanup()
+            if exc_type is None:
+                # Check for external command errors only when not already
+                # handling an exception.
+                self.check_errors()
 
 
 def quote(*args):
@@ -714,14 +805,18 @@ class ExternalCommandFailed(Exception):
     .. attribute:: command
 
        The :class:`ExternalCommand` object.
-
-    .. attribute:: returncode
-
-       The return code of the external command (an integer).
     """
 
     def __init__(self, command):
         self.command = command
-        self.returncode = command.returncode
-        error_message = "External command failed with exit code %s! (command: %s)"
-        super(ExternalCommandFailed, self).__init__(error_message % (command.returncode, quote(command.command_line)))
+        super(ExternalCommandFailed, self).__init__(command.error_message)
+
+    @property
+    def returncode(self):
+        """Shortcut for :attr:`ExternalCommand.returncode`."""
+        return self.command.returncode
+
+    @property
+    def error_message(self):
+        """Shortcut for :attr:`ExternalCommand.error_message`."""
+        return self.command.error_message
